package sim.app.species.ANN;

import ec.util.MersenneTwisterFast;
import sim.app.species.Creature;

//this is the entire neural network for one agent, it starts like a soup of somas and axons
//it creates and then specifies whats connected to what from reading the genome
//activates the whole network, contains input and output nodes
//responsible for neuron to neuron interaction
public abstract class Brain {

	private static double AXON_WEIGHT_MUT_RATE = 0.05;
	private static double NEW_SOMA_MUT_RATE = 0.005;
	private static double NEW_AXON_MUT_RATE = 0.005;

	// the compatibility weights
	private static double c1 = 1.0;
	private static double c2 = 1.0;
	private static double c3 = 0.4;
	private static double compThreshold = 1.0;

	private static final int numActions = Creature.numActions();
	private static int somaInnovNum = 0;
	private static int axonInnovNum = 0;

	private Axon[] allAxons;
	private Neuron[] neurons;
	private int numOutputs; // we'll keep track of this to make it easier to retrieve the outputs
	private SomaChromosome sChrome;
	private AxonChromosome aChrome;
	private MersenneTwisterFast random;

	public Brain(SomaChromosome nodes, AxonChromosome edges, MersenneTwisterFast random){

		int maxSomaNum, maxAxonNum;

		maxSomaNum = maxAxonNum = 0;

		this.sChrome = nodes;
		this.aChrome = edges;
		this.random = random;

		this.allAxons = new Axon[edges.length()];

		for(int i = 0; i < this.allAxons.length; i++){
			this.allAxons[i] = new Axon(edges.getGene(i).getID(), edges.getGene(i).getWeight());
			if(maxAxonNum < edges.getGene(i).getID()) maxAxonNum = edges.getGene(i).getID();
		}

		this.neurons = new Neuron[nodes.length()];
		this.numOutputs = 0;

		// loop through, and create a neuron for each soma
		for(int i = 0; i < nodes.length(); i++){

			SomaGene sGene = nodes.getGene(i);
			Soma soma = new Soma(sGene.getID());
			int numInCount, numOutCount;

			if(maxSomaNum < sGene.getID()) maxSomaNum = sGene.getID();

			numInCount = numOutCount = 0;

			// get a count of the number of input/output axons this soma has
			for(int j = 0; j < edges.length(); j++){

				AxonGene aGene = edges.getGene(j);

				// skip over it if it's disabled
				if(aGene.getDisabled()) continue;

				if(aGene.getEndID() == sGene.getID()) numInCount++;
				if(aGene.getStartID() == sGene.getID()) numOutCount++;
				
			}

			Axon[] incoming = new Axon[numInCount];
			Axon[] outgoing = new Axon[numOutCount];
			numInCount = numOutCount = 0;

			for(int j = 0; j < edges.length(); j++){

				AxonGene aGene = edges.getGene(j);

				// skip over it if it's disabled
				if(aGene.getDisabled()) continue;

				if(aGene.getStartID() == sGene.getID()) outgoing[numOutCount++] = this.allAxons[j];
				if(aGene.getEndID() == sGene.getID()) incoming[numInCount++] = this.allAxons[j];
				
			}

			switch(sGene.getType()){
				case SomaGene.INPUT:
					this.neurons[i] = new InputNeuron(soma, incoming, outgoing);
					break;
				case SomaGene.HIDDEN:
					this.neurons[i] = new HiddenNeuron(soma, incoming, outgoing);
					break;
				case SomaGene.OUTPUT:
					this.neurons[i] = new OutputNeuron(soma, incoming, outgoing);
					this.numOutputs++;
					break;
				case SomaGene.BIAS:
					this.neurons[i] = new BiasNeuron(soma, incoming, outgoing);
					break;
			}

		}

		Brain.updateSomaInnovNum(maxSomaNum+1);
		Brain.updateAxonInnovNum(maxAxonNum+1);

	}

	//input array must be the same size as the number of inputs to network, or bigger
	protected void setInput(double[] in){

		int count=0;

		for(int i = 0; i < this.neurons.length; i++){

			if(this.neurons[i] instanceof InputNeuron){
				((InputNeuron)neurons[i]).setInput(in[count++]);
			}

		}

	}
	//this has to know the number of output neurons, else i will have to do something dynamically
	//before returning the array
	protected double[] getOutput(){

		int count = 0;
		double[] out = new double[this.numOutputs];
		
		for(int i = 0; i < this.neurons.length; i++){

			if(this.neurons[i] instanceof OutputNeuron){
				out[count++]= this.neurons[i].soma.output;
			}
		}

		return out;

	}

	//moves all signals one time tick
	protected void activate(){

		for(int i = 0; i < this.neurons.length; i++){
			this.neurons[i].fire();	
		}

		for(int i = 0; i < this.neurons.length; i++){
			if(this.neurons[i].soma.isActive()){
				this.neurons[i].setOutAxonsActive();
			}
		}

		// sets all inputs at time (t-1) to the outputs at time t
		for(int i = 0; i < this.allAxons.length; i++){
			this.allAxons[i].fire();
		}

	}

	/* Uses this and that to create a new offspring
	 * The crossover takes place like described in the NEAT paper
	 * Returns either an AdamBrain or DavidBrain, depending on what types are passed to it
	 **/
	public Brain createOffspring(Brain that){

		Brain newBrain;
		SomaChromosome sChromeNew = new SomaChromosome();
		AxonChromosome aChromeNew = new AxonChromosome();
		SomaGene thisSomaGene, thatSomaGene, newSomaGene;
		AxonGene thisAxonGene, thatAxonGene, newAxonGene;
		int thisIndex, thatIndex;

		// create the soma chromosome by just getting each unique soma gene from the parent brains
		thisIndex = thatIndex = 0;
		while(thisIndex < this.sChrome.length() && thatIndex < that.sChrome.length()){

			thisSomaGene = this.sChrome.getGene(thisIndex);
			thatSomaGene = that.sChrome.getGene(thatIndex);

			if(thisSomaGene.getID() == thatSomaGene.getID()){

				newSomaGene = new SomaGene(thisSomaGene.getID(), thisSomaGene.getType());
				thisIndex++;
				thatIndex++;

			}else if(thisSomaGene.getID() < thatSomaGene.getID()){

				newSomaGene = new SomaGene(thisSomaGene.getID(), thisSomaGene.getType());
				thisIndex++;

			}else{

				newSomaGene = new SomaGene(thatSomaGene.getID(), thatSomaGene.getType());
				thatIndex++;

			}

			sChromeNew.addGene(newSomaGene);

		}

		/* NOTE: ONLY ONE OF THE TWO FOR LOOPS WILL GET EXECUTED */
		// add the soma genes at the end of this brain
		for(int i = thisIndex; i < this.sChrome.length(); i++){
			thisSomaGene = this.sChrome.getGene(i);
			newSomaGene = new SomaGene(thisSomaGene.getID(), thisSomaGene.getType());
			sChromeNew.addGene(newSomaGene);
		}

		// add the soma genes at the end of that brain
		for(int i = thatIndex; i < that.sChrome.length(); i++){
			thatSomaGene = that.sChrome.getGene(i);
			newSomaGene = new SomaGene(thatSomaGene.getID(), thatSomaGene.getType());
			sChromeNew.addGene(newSomaGene);
		}

		/* create the axon chromosome by getting all axons from the two parents
		 * if each parent has the axon, then average the weights
		 **/
		thisIndex = thatIndex = 0;
		while(thisIndex < this.aChrome.length() && thatIndex < that.aChrome.length()){

			thisAxonGene = this.aChrome.getGene(thisIndex);
			thatAxonGene = that.aChrome.getGene(thatIndex);

			if(thisAxonGene.getID() == thatAxonGene.getID()){

				double w = (thisAxonGene.getWeight()+thatAxonGene.getWeight())/2.0;
				boolean dis = false;

				if(thisAxonGene.getDisabled() == thatAxonGene.getDisabled()){
					dis = thisAxonGene.getDisabled();
				}else{
					dis = this.random.nextBoolean();
				}

				newAxonGene = new AxonGene(thisAxonGene.getID(), thisAxonGene.getStartID(), thisAxonGene.getEndID(), w, dis);
				thisIndex++;
				thatIndex++;

			}else if(thisAxonGene.getID() < thatAxonGene.getID()){

				newAxonGene = new AxonGene(thisAxonGene.getID(), thisAxonGene.getStartID(), thisAxonGene.getEndID(), thisAxonGene.getWeight(), thisAxonGene.getDisabled());
				thisIndex++;

			}else{

				newAxonGene = new AxonGene(thatAxonGene.getID(), thatAxonGene.getStartID(), thatAxonGene.getEndID(), thatAxonGene.getWeight(), thatAxonGene.getDisabled());
				thatIndex++;

			}

			aChromeNew.addGene(newAxonGene);

		}

		/* NOTE: ONLY ONE OF THE TWO FOR LOOPS WILL GET EXECUTED */
		// add the axon genes at the end of this brain
		for(int i = thisIndex; i < this.aChrome.length(); i++){
			thisAxonGene = this.aChrome.getGene(i);
			newAxonGene = new AxonGene(thisAxonGene.getID(), thisAxonGene.getStartID(), thisAxonGene.getEndID(), thisAxonGene.getWeight(), thisAxonGene.getDisabled());
			aChromeNew.addGene(newAxonGene);
		}

		// add the axon genes at the end of that brain
		for(int i = thatIndex; i < that.aChrome.length(); i++){
			thatAxonGene = that.aChrome.getGene(i);
			newAxonGene = new AxonGene(thatAxonGene.getID(), thatAxonGene.getStartID(), thatAxonGene.getEndID(), thatAxonGene.getWeight(), thatAxonGene.getDisabled());
			aChromeNew.addGene(newAxonGene);
		}

		// recombination is complete, now mutate
		this.mutateAxonWeights(aChromeNew);
		this.mutateNewSoma(sChromeNew, aChromeNew);
		this.mutateNewAxon(sChromeNew, aChromeNew);

		if(this instanceof AdamBrain){
			return new AdamBrain(sChromeNew, aChromeNew, this.random);
		}else{
			return new DavidBrain(sChromeNew, aChromeNew, this.random);
		}

	}

	private static void updateSomaInnovNum(int n){

		if(Brain.somaInnovNum < n) Brain.somaInnovNum = n;

	}

	private static void updateAxonInnovNum(int n){

		if(Brain.axonInnovNum < n) Brain.axonInnovNum = n;

	}

	private void mutateAxonWeights(AxonChromosome aChr){

		double w;

		for(int i = 0; i < aChr.length(); i++){

			if(this.random.nextDouble() < Brain.AXON_WEIGHT_MUT_RATE){
				double d = this.random.nextGaussian();
				aChr.getGene(i).addWeight(d);
			}

		}

	}

	// create a new soma by inserting it inbetween two already connected somas
	private void mutateNewSoma(SomaChromosome sChr, AxonChromosome aChr){

		if(this.random.nextDouble() < Brain.NEW_SOMA_MUT_RATE){

			int index = this.random.nextInt(aChr.length());
			SomaGene newSomaGene;
			AxonGene oldAxonGene, newAxonGene1, newAxonGene2;

			// find an axon that is not disabled
			while(aChr.getGene(index).getDisabled()){
				index = this.random.nextInt(aChr.length());
			}

			oldAxonGene = aChr.getGene(index);

			// disable the original axon
			oldAxonGene.setDisabled(true);

			// increment the InnovNums when doing this!
			newSomaGene = new SomaGene(this.somaInnovNum++, SomaGene.HIDDEN);
			newAxonGene1 = new AxonGene(this.axonInnovNum++, oldAxonGene.getStartID(), newSomaGene.getID(), oldAxonGene.getWeight());
			newAxonGene2 = new AxonGene(this.axonInnovNum++, newSomaGene.getID(), oldAxonGene.getEndID(), oldAxonGene.getWeight());

			sChr.addGene(newSomaGene);
			aChr.addGene(newAxonGene1);
			aChr.addGene(newAxonGene2);

		}

	}

	/* create a new axon by randomly generating one and inserting it into the network
	 *
	 * The new axon will never be pointing to an input neuron
	 * Never duplicate axons
	 **/
	private void mutateNewAxon(SomaChromosome sChr, AxonChromosome aChr){

		if(this.random.nextDouble() < Brain.NEW_AXON_MUT_RATE){

			int startIndex, endIndex;
			int startID, endID;
			int tries = 0;
			boolean isGood;
			AxonGene newAxonGene;

			while(tries < 100){

				startIndex = this.random.nextInt(sChr.length());
				endIndex = this.random.nextInt(sChr.length());

				// find a non-input node
				while(sChr.getGene(endIndex).getType() == SomaGene.INPUT || sChr.getGene(endIndex).getType() == SomaGene.BIAS){
					endIndex = this.random.nextInt(sChr.length());
				}

				startID = sChr.getGene(startIndex).getID();
				endID = sChr.getGene(endIndex).getID();

				isGood = true;

				// make sure link is does not already exist
				for(int i = 0; i < aChr.length(); i++){
					if(aChr.getGene(i).getStartID() == startID && aChr.getGene(i).getEndID() == endID){
						isGood = false;
						break;
					} 
				}

				if(isGood){
					newAxonGene = new AxonGene(Brain.axonInnovNum++, startID, endID, this.random);
					aChr.addGene(newAxonGene);
					break;
				}else{
					tries++;
				}

			}

		}

	}

	/* returns true if the two brains are compatible for crossover
	 **/
	public boolean isCompatible(Brain that){

		double distance, avgWeightDiff;
		int somaExcess, axonExcess, totalExcess;
		int disjointAxons, disjointSomas, totalDisjoint;
		int countSameAxons=0,countSameSomas=0;;
		int thisIndex, thatIndex;
		SomaGene thisSomaGene, thatSomaGene;
		AxonGene thisAxonGene, thatAxonGene;

		//find number of disjoint axons as well as sum the weights of same ones
		thisIndex = thatIndex = 0;
		avgWeightDiff = 0.0;
		disjointAxons = Math.min(this.aChrome.length(), that.aChrome.length());

		while(thisIndex < this.aChrome.length() && thatIndex < that.aChrome.length()){

			thisAxonGene = this.aChrome.getGene(thisIndex);
			thatAxonGene = that.aChrome.getGene(thatIndex);

			if(thisAxonGene.getID() == thatAxonGene.getID()){

				countSameAxons++;
				disjointAxons--;

				avgWeightDiff += Math.abs(thisAxonGene.getWeight() - thatAxonGene.getWeight());

				thisIndex++;
				thatIndex++;

			}else if(thisAxonGene.getID() < thatAxonGene.getID()){
				thisIndex++;
			}else{
				thatIndex++;
			}

		}
		
		axonExcess = Math.max(this.aChrome.length() - thisIndex, that.aChrome.length() - thatIndex);

		// go through the soma gense
		thisIndex = thatIndex = 0;
		disjointSomas = Math.min(this.sChrome.length(), that.sChrome.length());

		while(thisIndex < this.sChrome.length() && thatIndex < that.sChrome.length()){

			thisSomaGene = this.sChrome.getGene(thisIndex);
			thatSomaGene = that.sChrome.getGene(thatIndex);

			if(thisSomaGene.getID() == thatSomaGene.getID()){

				countSameSomas++;
				disjointSomas--;
				thisIndex++;
				thatIndex++;

			}else if(thisSomaGene.getID() < thatSomaGene.getID()){
				thisIndex++;	

			}else{
				thatIndex++;
			}

		}

		somaExcess = Math.max(this.sChrome.length() - thisIndex, that.sChrome.length() - thatIndex);

		totalDisjoint = disjointAxons + disjointSomas;

		if(countSameAxons==0){
			avgWeightDiff = 0.0;			
		}else{
			avgWeightDiff /= countSameAxons;
		}

		totalExcess = somaExcess + axonExcess;

		distance = Brain.c1 * totalExcess + Brain.c2 * totalDisjoint + Brain.c3 * avgWeightDiff;

		return distance < Brain.compThreshold;

	}

	public static void addToCompThreshold(double d){

		Brain.compThreshold += d;

		if(Brain.compThreshold < 0.0) Brain.compThreshold = 0.0;

	}

	public static double getCompatibilityThreshold(){
		return Brain.compThreshold;
	}

	public static void setCompatibilityThreshold(double d){
		Brain.compThreshold = d;
	}

	public static void setCompatibility1(double d){
		Brain.c1 = d;
	}

	public static void setCompatibility2(double d){
		Brain.c2 = d;
	}

	public static void setCompatibility3(double d){
		Brain.c3 = d;
	}

	public static void setAxonWeightMutRate(double d){
		Brain.AXON_WEIGHT_MUT_RATE = d;
	}

	public static void setNewSomaMutRate(double d){
		Brain.NEW_SOMA_MUT_RATE = d;
	}

	public static void setNewAxonMutRate(double d){
		Brain.NEW_AXON_MUT_RATE = d;
	}

	protected static int getNumActions(){
		return Brain.numActions;
	}

	// abstract functions
	public abstract double[] activate(double[] inv);

	public int getSomaChromeLen(){
		return this.sChrome.length();
	}

	public int getAxonChromeLen(){
		return this.aChrome.length();
	}

	public String toString(){

		String ret = new String();

		ret = "SomaCount: " + this.sChrome.length() + "\n";
		ret += "AxonCount: " + this.aChrome.length() + "\n";

		for(int i = 0; i < this.sChrome.length(); i++){
			ret += sChrome.getGene(i).toString() + "\n";
		}

		for(int i = 0; i < this.aChrome.length(); i++){
			ret += aChrome.getGene(i).toString() + "\n";
		}

		return ret;

	}

}
