package sim.app.species;

import sim.engine.*;
import sim.portrayal.*;
import sim.util.*;
import java.awt.*;
import java.awt.geom.*;
import java.util.*;

public abstract class Entity extends SimplePortrayal2D implements Steppable{

	// this will most likely never change
	// only two possible values for this: 3 or 6
	protected static final int NUM_BITS_IN_LOOK = 3;

	protected MutableDouble2D heading, position;
	private boolean alive;
	private boolean[] look;
	private double speed;
	private int energy, lifeSpan, age;
	private double diameter, radius;
	private double senseDistance; // the distance the Entity can "see"
	private boolean drawSensor, drawHeading;

	// these affect the way the entities are drawn
	private Color color;
	private boolean showLook;

	// not sure if keeping the stoppable object in here is a good idea
	protected Stoppable stoppable;

	public Entity(MutableDouble2D initPos, MutableDouble2D initHead){

		this.alive = true;
		this.age = 0;

		this.position = new MutableDouble2D(initPos);
		this.heading = new MutableDouble2D(initHead);
		this.drawSensor = false;
		this.drawHeading = false;

	}

	public Entity(){ this(new MutableDouble2D(), new MutableDouble2D()); }

	// figure out the color of this Entity, based on its look
	protected void resetColor(){

		int r, g, b;

		r = g = b = 0;

		if(Entity.NUM_BITS_IN_LOOK == 3){

			if(look[0]) r = 192;
			if(look[1]) g = 192;
			if(look[2]) b = 192;

		}else if(Entity.NUM_BITS_IN_LOOK == 6){

			if(look[0]) r += 64;
			if(look[1]) r += 128;
			if(look[2]) g += 64;
			if(look[3]) g += 128;
			if(look[4]) b += 64;
			if(look[5]) b += 128;

		}

		this.color = new Color(r, g, b);

	}

	/* 
	 **/
	public void draw(Object object, Graphics2D graphics, DrawInfo2D info){

		// get out of here if the thing is dead
		if(!this.alive) return;

		double width, height;
		int x, y, w, h;

		// draw a yellow circle around the Creature if it just mated
		if(this instanceof Creature && ((Creature)this).justMated > 0){

			width = info.draw.width * this.diameter * 3.0;
			height = info.draw.height * this.diameter * 3.0;
			x = (int)(info.draw.x - width / 2.0);
			y = (int)(info.draw.y - height / 2.0);
			w = (int)(width);
			h = (int)(height);

			// draw the sensor centered on the origin
			graphics.setColor(Color.yellow);
			graphics.fillOval(x, y, w, h);

		}

		// draw a green circle around the Creature if it's one of the best moving food eaters
		if(this instanceof Creature && ((Creature)this).isBestMovingFoodEater){

			width = info.draw.width * this.diameter * 3.0;
			height = info.draw.height * this.diameter * 3.0;
			x = (int)(info.draw.x - width / 2.0);
			y = (int)(info.draw.y - height / 2.0);
			w = (int)(width);
			h = (int)(height);

			// draw the sensor centered on the origin
			graphics.setColor(Color.green);
			graphics.drawOval(x, y, w, h);

		}

		// draw a red circle around the Creature if it's one of the best stationary food eaters
		if(this instanceof Creature && ((Creature)this).isBestStationaryFoodEater){

			width = info.draw.width * this.diameter * 3.0;
			height = info.draw.height * this.diameter * 3.0;
			x = (int)(info.draw.x - width / 2.0);
			y = (int)(info.draw.y - height / 2.0);
			w = (int)(width);
			h = (int)(height);

			// draw the sensor centered on the origin
			graphics.setColor(Color.red);
			graphics.drawOval(x, y, w, h);

		}

		// values for the Entity
		width = info.draw.width * this.diameter;
		height = info.draw.height * this.diameter;
		x = (int)(info.draw.x - width / 2.0);
		y = (int)(info.draw.y - height / 2.0);
		w = (int)(width);
		h = (int)(height);

		// draw the AV centered on the origin
		if(this instanceof Creature){
			graphics.setColor(((Creature)this).speciesColor());
		}else{
			graphics.setColor(this.color);
		}
		graphics.fillOval(x, y, w, h);

		if(this.drawSensor){

			width = info.draw.width * this.senseDistance * 2.0;
			height = info.draw.height * this.senseDistance * 2.0;
			x = (int)(info.draw.x - width / 2.0);
			y = (int)(info.draw.y - height / 2.0);
			w = (int)(width);
			h = (int)(height);

			// draw the sensor centered on the origin
			graphics.setColor(Color.red);
			graphics.drawOval(x, y, w, h);

			if(this instanceof Creature){

				int xline = (int)(info.draw.x + this.heading.y * (info.draw.width * this.senseDistance));
				int yline = (int)(info.draw.y + (-this.heading.x) * (info.draw.height * this.senseDistance));
				MutableDouble2D segment = new MutableDouble2D(xline-info.draw.x, yline-info.draw.y);
				final double angle = Math.PI/Creature.numVisionDivisions();

				for(int i = 0; i <= Creature.numVisionDivisions(); i++){

					graphics.drawLine((int)info.draw.x, (int)info.draw.y, (int)(segment.x + info.draw.x), (int)(segment.y + info.draw.y));

					segment.rotate(angle);

				}

			}

		}

		if(this.drawHeading){

			// values for the direction
			final int xline, yline;

			xline = (int)(info.draw.x + this.heading.x * (info.draw.width * this.radius));
			yline = (int)(info.draw.y + this.heading.y * (info.draw.height * this.radius));

			// draw its direction
			graphics.setColor(Color.white);
			graphics.drawLine((int)info.draw.x, (int)info.draw.y, xline, yline);

		}

		// values for the squares
		if(this.showLook){

			final int drawSquareLenX = (int)(this.diameter*info.draw.width/4.0);
			final int drawSquareLenY = (int)(this.diameter*info.draw.height/4.0);

			if(Entity.NUM_BITS_IN_LOOK == 3){

				if(this.look[0]) graphics.setColor(Color.black);
				else             graphics.setColor(Color.white);
				graphics.fillRect((int)(info.draw.x-1.5*drawSquareLenX), (int)(info.draw.y-0.5*drawSquareLenY), drawSquareLenX, drawSquareLenY);

				if(this.look[1]) graphics.setColor(Color.black);
				else             graphics.setColor(Color.white);
				graphics.fillRect((int)(info.draw.x-0.5*drawSquareLenX), (int)(info.draw.y-0.5*drawSquareLenY), drawSquareLenX, drawSquareLenY);

				if(this.look[2]) graphics.setColor(Color.black);
				else             graphics.setColor(Color.white);
				graphics.fillRect((int)(info.draw.x+0.5*drawSquareLenX), (int)(info.draw.y-0.5*drawSquareLenY), drawSquareLenX, drawSquareLenY);

			}else if(Entity.NUM_BITS_IN_LOOK == 6){

				if(this.look[0]) graphics.setColor(Color.black);
				else             graphics.setColor(Color.white);
				graphics.fillRect((int)(info.draw.x-drawSquareLenX), (int)(info.draw.y-1.5*drawSquareLenY), drawSquareLenX, drawSquareLenY);

				if(this.look[1]) graphics.setColor(Color.black);
				else             graphics.setColor(Color.white);
				graphics.fillRect((int)info.draw.x, (int)(info.draw.y-1.5*drawSquareLenY), drawSquareLenX, drawSquareLenY);

				if(this.look[2]) graphics.setColor(Color.black);
				else             graphics.setColor(Color.white);
				graphics.fillRect((int)(info.draw.x-1.5*drawSquareLenX), (int)(info.draw.y-0.5*drawSquareLenY), drawSquareLenX, drawSquareLenY);

				if(this.look[3]) graphics.setColor(Color.black);
				else             graphics.setColor(Color.white);
				graphics.fillRect((int)(info.draw.x+0.5*drawSquareLenX), (int)(info.draw.y-0.5*drawSquareLenY), drawSquareLenX, drawSquareLenY);

				if(this.look[4]) graphics.setColor(Color.black);
				else             graphics.setColor(Color.white);
				graphics.fillRect((int)(info.draw.x-drawSquareLenX), (int)(info.draw.y+0.5*drawSquareLenY), drawSquareLenX, drawSquareLenY);

				if(this.look[5]) graphics.setColor(Color.black);
				else             graphics.setColor(Color.white);
				graphics.fillRect((int)info.draw.x, (int)(info.draw.y+0.5*drawSquareLenY), drawSquareLenX, drawSquareLenY);

			}

		}

	}

	/* Each Entity in the simulation is a circle.
	 * This code was taken from the Ball.java file from Tutorial 5
	 *   as the authors of the tutorial should know how to write 
	 *   the most efficient code for this function.
	 **/	
	public boolean hitObject(Object object, DrawInfo2D range){

		final double SLOP = 1.0;  // need a little extra diameter to hit circles
		final double width = range.draw.width * this.radius;
		final double height = range.draw.height * this.diameter;

		Ellipse2D.Double ellipse = new Ellipse2D.Double( 
			range.draw.x-width/2-SLOP, 
			range.draw.y-height/2-SLOP,
			width+SLOP*2,
			height+SLOP*2 );
		return ( ellipse.intersects( range.clip.x, range.clip.y, range.clip.width, range.clip.height ) );
	
	}

	// updates the position of the entity based on its current heading and speed
	public void updatePosition(){

		this.position.addIn(this.heading.x * this.speed, this.heading.y * this.speed);

	}

	// updates the heading of the entity if it is hitting a wall
	public void updateHeadingHitWall(int simWidth, int simHeight){

		/* reverse direction if a wall is encountered
		 **/
		if((this.position.x - this.radius <= 0.0 && this.heading.x < 0.0) || 
		   (this.position.x + this.radius >= simWidth && this.heading.x > 0.0)){
			this.heading.x *= -1.0;
		}

		if((this.position.y - this.radius <= 0.0 && this.heading.y < 0.0) || 
		   (this.position.y + this.radius >= simHeight && this.heading.y > 0.0)){
			this.heading.y *= -1.0;
		}

	}

	// returns true if the entity is hitting a wall
	public boolean hittingWall(int simWidth, int simHeight){

		return (this.position.x - this.radius <= 0.0 && this.heading.x < 0.0) || 
		       (this.position.x + this.radius >= simWidth && this.heading.x > 0.0) ||
		       (this.position.y - this.radius <= 0.0 && this.heading.y < 0.0) || 
		       (this.position.y + this.radius >= simHeight && this.heading.y > 0.0);

	}

	/* returns true if both entities are alive and they are touching
	 **/
	public boolean isTouching(Entity that){

		if(!this.alive || !that.alive) return false;

		return (this.position.distance(that.position) <= this.radius + that.radius);

	}

	// simply adds in the value to the entity's energy
	public void addEnergy(int n){ this.energy += n; }

	// setters and getters
	public boolean isAlive(){ return this.alive; }
	public void setAlive(boolean b){ this.alive = b; }
	public int getEnergy(){ return this.energy; }
	public void setEnergy(int n){ this.energy = n; }
	public boolean getShowLook(){ return this.showLook; }
	public void setShowLook(boolean b){ this.showLook = b; }
	public double getDiameter(){ return this.diameter; }
	public double getRadius(){ return this.radius; }
	public void setDiameter(double d){ this.diameter = d; this.radius = d/2.0; }
	public void setSenseDistance(double d){ this.senseDistance = d; }
	public double senseDistance(){ return this.senseDistance; }
	public double getSpeed(){ return this.speed; }
	public void setSpeed(double d){ this.speed = d; }
	public boolean[] getLook(){ return this.look; }
	public void setLifeSpan(int n){ this.lifeSpan = n; }
	public int getLifeSpan(){ return this.lifeSpan; }
	public void incrementAge(){ this.age++; }
	public int getAge(){ return this.age; }
	public void setAge(int n){ this.age = n; }
	public void setDrawSensor(boolean b){ this.drawSensor = b; }
	public boolean getDrawSensor(){ return this.drawSensor; }
	public void setDrawHeading(boolean b){ this.drawHeading = b; }
	public boolean getDrawHeading(){ return this.drawHeading; }

	// this function throws a new exception if the array is not of size 6
	public void setLook(boolean[] b){

		if(b.length != Entity.NUM_BITS_IN_LOOK){
			throw new RuntimeException("The look for an Entity must be a boolean array of size " + Entity.NUM_BITS_IN_LOOK);
		}

		this.look = new boolean[b.length];

		for(int i = 0; i < look.length; i++) this.look[i] = b[i];

		// now that the look is updated, don't forget to reset its color
		this.resetColor();

	}

}
